# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for JAX stubs."""

from typing import Any, NewType

from absl.testing import absltest
import jax.numpy as jnp
from tensor_annotations import axes
from tensor_annotations.jax import Array0
from tensor_annotations.jax import Array1
from tensor_annotations.jax import Array2
from tensor_annotations.tests import utils


A1 = NewType('A1', axes.Axis)
A2 = NewType('A2', axes.Axis)

# It's less than ideal that we have to repeat imports etc. here for pytype, but
# this seems like the best balance between readability and complexity.
_PREAMBLE = """
from typing import NewType

import jax.numpy as jnp
from tensor_annotations import axes
from tensor_annotations.jax import Array0, Array1, Array2

A1 = NewType('A1', axes.Axis)
A2 = NewType('A2', axes.Axis)
"""


class JAXStubTests(absltest.TestCase):

  def testTranspose_InferredShapeMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Array2[A1, A2] = jnp.zeros((1, 2))
      y = jnp.transpose(x)

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.y, y.shape)

  def testUnaryOperator_ReturnCustomType(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Array1[A1] = jnp.zeros((1,))
      # Let's just test a representative subset.
      a = jnp.abs(x)  # pylint: disable=unused-variable
      b = jnp.sin(x)  # pylint: disable=unused-variable
      c = jnp.floor(x)  # pylint: disable=unused-variable
      d = jnp.ones_like(x)  # pylint: disable=unused-variable
      e = jnp.round(x)  # pylint: disable=unused-variable
      f = jnp.sign(x)  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    expected = 'Array1[A1]'
    self.assertEqual(inferred.a, expected)
    self.assertEqual(inferred.b, expected)
    self.assertEqual(inferred.c, expected)
    self.assertEqual(inferred.d, expected)
    self.assertEqual(inferred.e, expected)
    self.assertEqual(inferred.f, expected)

  def testZerosOnes_ReturnsCorrectShape(self):
    with utils.SaveCodeAsString() as code_saver:
      a = jnp.zeros(())  # pylint: disable=unused-variable
      b = jnp.ones(())  # pylint: disable=unused-variable
      c = jnp.zeros((1,))  # pylint: disable=unused-variable
      d = jnp.ones((1,))  # pylint: disable=unused-variable
      e = jnp.zeros((1, 1))  # pylint: disable=unused-variable
      f = jnp.ones((1, 1))  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.a, 'Array0')
    self.assertEqual(inferred.b, 'Array0')
    self.assertEqual(inferred.c, 'Array1')
    self.assertEqual(inferred.d, 'Array1')
    self.assertEqual(inferred.e, 'Array2')
    self.assertEqual(inferred.f, 'Array2')

  def testSum_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Array2[A1, A2] = jnp.zeros((1, 2))
      y1 = jnp.sum(x, axis=0)
      y2 = jnp.sum(x, axis=1)
      y3 = jnp.sum(x, axis=(0, 1))
      y4 = jnp.sum(x)

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.y1, y1.shape)
    self.assertEqual(inferred.y2, y2.shape)
    self.assertEqual(inferred.y3, y3.shape)
    self.assertEqual(inferred.y4, y4.shape)

  def testSumKeepdimsTrue_ReturnsAny(self):
    # We haven't got around to making stubs for keepdims=True yet;
    # make sure the type reflects that.
    with utils.SaveCodeAsString() as code_saver:
      x: Array1[A1] = jnp.zeros((1,))
      a = jnp.sum(x, axis=0, keepdims=True)  # pylint: disable=unused-variable
      b = jnp.sum(x, keepdims=True)  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.a, 'Any')
    self.assertEqual(inferred.b, 'Any')

  def testTensorAdd_ReturnsCustomType(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Array1[A1] = jnp.zeros((1,))
      a = x + 1  # pylint: disable=unused-variable
      b = x + x  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual('Array1[A1]', inferred.a)
    self.assertEqual('Array1[A1]', inferred.b)

  def testTensorUnaryOp_ReturnsCorrectTypeAndShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x1: Array0 = jnp.zeros(())
      y1 = abs(x1)  # pylint: disable=unused-variable
      y2 = -x1  # pylint: disable=unused-variable
      x2: Array1[A1] = jnp.zeros((1,))
      y3 = abs(x2)  # pylint: disable=unused-variable
      y4 = -x2  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual('Array0', inferred.y1)
    self.assertEqual('Array0', inferred.y2)
    self.assertEqual('Array1[A1]', inferred.y3)
    self.assertEqual('Array1[A1]', inferred.y4)

  def testBinaryOpWithScalar_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Array2[A1, A2] = jnp.zeros((1, 2))
      y1 = x + 1.0
      y2 = x - 1.0
      y3 = x / 1.0
      y4 = x * 1.0

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(y1.shape, inferred.y1)
    self.assertEqual(y2.shape, inferred.y2)
    self.assertEqual(y3.shape, inferred.y3)
    self.assertEqual(y4.shape, inferred.y4)

  def testBinaryOpWithBroadcast_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      a: Array2[A1, A2] = jnp.zeros((1, 2))
      b: Array1[A2] = jnp.zeros((2,))
      y1 = a + b
      y2 = a - b
      y3 = a / b
      y4 = a * b

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(y1.shape, inferred.y1)
    self.assertEqual(y2.shape, inferred.y2)
    self.assertEqual(y3.shape, inferred.y3)
    self.assertEqual(y4.shape, inferred.y4)

  def testBinaryOpWithSameShape_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      a: Array2[A1, A2] = jnp.zeros((1, 2))
      b: Array2[A1, A2] = jnp.zeros((1, 2))
      y1 = a + b
      y2 = a - b
      y3 = a / b
      y4 = a * b

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(y1.shape, inferred.y1)
    self.assertEqual(y2.shape, inferred.y2)
    self.assertEqual(y3.shape, inferred.y3)
    self.assertEqual(y4.shape, inferred.y4)

  def testShapeAttribute_HasCorrectLength(self):
    with utils.SaveCodeAsString() as code_saver:
      x0 = jnp.zeros(())
      x1 = jnp.zeros((1,))
      x2 = jnp.zeros((1, 2))
      x3 = jnp.zeros((1, 2, 3))
      x4 = jnp.zeros((1, 2, 3, 4))
      x0_shape = x0.shape   # pylint: disable=unused-variable
      x1_shape = x1.shape   # pylint: disable=unused-variable
      x2_shape = x2.shape   # pylint: disable=unused-variable
      x3_shape = x3.shape   # pylint: disable=unused-variable
      x4_shape = x4.shape   # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual(x0_shape, ())
    self.assertEqual(x1_shape, (1,))
    self.assertEqual(x2_shape, (1, 2))
    self.assertEqual(x3_shape, (1, 2, 3))
    self.assertEqual(x4_shape, (1, 2, 3, 4))
    self.assertEqual('Tuple[()]', inferred.x0_shape)
    self.assertEqual('Tuple[int]', inferred.x1_shape)
    self.assertEqual('Tuple[int, int]', inferred.x2_shape)
    self.assertEqual('Tuple[int, int, int]', inferred.x3_shape)
    self.assertEqual('Tuple[int, int, int, int]', inferred.x4_shape)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests to confirm that Pytype helpers function correctly."""

import textwrap
import unittest

from tensor_annotations.tests import utils


_PREAMBLE = """
from typing import NewType

from tensor_annotations.jax import Array2
from tensor_annotations import axes

A1 = NewType('A1', axes.Axis)
A2 = NewType('A2', axes.Axis)
"""


class PytypeTests(unittest.TestCase):

  def testSimpleCorrectExample_PassesPytype(self):
    code = """
      def foo(x: Array2[A1, A2]):
        pass
      x: Array2[A1, A2] = Array2()
      foo(x)
    """
    code = _PREAMBLE + textwrap.dedent(code)
    utils.assert_pytype_succeeds(code)

  def testSimpleIncorrectExample_FailsPytype(self):
    code = """
      def foo(x: Array2[A1, A2]):
        pass
      x: Array2[A2, A1] = Array2()
      foo(x)
    """
    code = _PREAMBLE + textwrap.dedent(code)
    utils.assert_pytype_fails(code)


if __name__ == '__main__':
  unittest.main()



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Custom tensor classes for JAX supporting shape parameterisation.

Note that these should only be used for the purposes of type annotation and
should never be instantiated. (Certain IDEs may also use these for
autocompletion, too.)

Type annotations for these classes are maintained in a separate stubs file,
`tensorflow.pyi`.
"""

# LINT.IfChange

from typing import Any, Generic, TypeVar

from tensor_annotations import axes

A1 = TypeVar('A1', bound=axes.Axis)
A2 = TypeVar('A2', bound=axes.Axis)
A3 = TypeVar('A3', bound=axes.Axis)
A4 = TypeVar('A4', bound=axes.Axis)
A5 = TypeVar('A5', bound=axes.Axis)
A6 = TypeVar('A6', bound=axes.Axis)
A7 = TypeVar('A7', bound=axes.Axis)
A8 = TypeVar('A8', bound=axes.Axis)


class _TensorBase:
  """Base class for TensorN classes containing common methods and attributes."""

  def __new__(cls, *args, **kwargs):
    raise TypeError('tensor_annotations tensors should not be instantiated')

  # These are necessary so that type checkers know we have these methods.
  __abs__: Any
  __add__: Any
  __add__: Any
  __floordiv__: Any
  __ge__: Any
  __gt__: Any
  __le__: Any
  __lt__: Any
  __mul__: Any
  __neg__: Any
  __neg__: Any
  __pos__: Any
  __pow__: Any
  __rmul__: Any
  __sub__: Any
  __truediv__: Any
  shape: Any


class Tensor0(_TensorBase):
  """A scalar - produced by e.g. tf.reduce_sum(tf.zeros((2, 3)))."""
  pass


class Tensor1(Generic[A1], _TensorBase):
  """A tensor of rank 1."""
  pass


class Tensor2(Generic[A1, A2], _TensorBase):
  """A tensor of rank 2."""
  pass


class Tensor3(Generic[A1, A2, A3], _TensorBase):
  """A tensor of rank 3."""
  pass


class Tensor4(Generic[A1, A2, A3, A4], _TensorBase):
  """A tensor of rank 4."""
  pass


class Tensor5(Generic[A1, A2, A3, A4, A5], _TensorBase):
  """A tensor of rank 5."""
  pass


class Tensor6(Generic[A1, A2, A3, A4, A5, A6], _TensorBase):
  """A tensor of rank 6."""
  pass


class Tensor7(Generic[A1, A2, A3, A4, A5, A6, A7], _TensorBase):
  """A tensor of rank 7."""
  pass


class Tensor8(Generic[A1, A2, A3, A4, A5, A6, A7, A8], _TensorBase):
  """A tensor of rank 8."""
  pass

# LINT.ThenChange(tensorflow.pyi)



"""Decorators for confirming shapes are correct at runtime."""

import copy
import functools
import inspect
import textwrap
import typing
from typing import Any, Dict, Mapping, Tuple

import tensor_annotations.jax as tjax
import tensor_annotations.tensorflow as ttf
import tree


# TODO: Replace this with something easier to maintain.
_TYPES_TO_CHECK = [
    ttf.Tensor0,
    ttf.Tensor1,
    ttf.Tensor2,
    ttf.Tensor3,
    ttf.Tensor4,
    ttf.Tensor5,
    ttf.Tensor6,
    ttf.Tensor7,
    tjax.Array0,
    tjax.Array1,
    tjax.Array2,
    tjax.Array3,
    tjax.Array4,
]


def _is_tensor_type(t):
  if not hasattr(t, '__origin__'):
    # It's not a generic type, so can't be one of the types we care about.
    return False
  # If t is Tensor1[Height], then t.__origin__ == Tensor1.
  if not any(t.__origin__ is generic_type
             for generic_type in _TYPES_TO_CHECK):
    return False
  return True


def verify_runtime_ranks_of_args_and_return(func):
  """Decorator that verifies ranks of arguments and return are correct.

  For example, if an argument `x` is annotated as having type
  `Tensor2[Height, Width]`, we verify that `len(x.shape) == 2`.

  Note that nested argument and return types are not verified.
  For example, if the return type is `Tuple[int, Tensor2[Height, Width]]`,
  we give up and do no checks.

  Args:
    func: The function to decorate.

  Raises:
    TypeError: If rank of return type or any argument is incorrect.

  Returns:
    Decorated function.
  """

  return functools.partial(_verify_runtime_args_and_return_ranks, func)


def _verify_runtime_args_and_return_ranks(func, *args, **kwargs):
  """Main implementation of verify_runtime_args_and_return_ranks."""
  sig: inspect.Signature = inspect.signature(func)

  # ===== Verify args. =====

  bound_arguments: inspect.BoundArguments = sig.bind(*args, **kwargs)
  # Do some args have default values, but the arg was not specified?
  # If so, set the args to their default values.
  bound_arguments.apply_defaults()
  arg_value_by_name: Dict[str, Any] = bound_arguments.arguments
  arg_signature_by_name: Mapping[str, inspect.Parameter] = sig.parameters

  # Note: we iterate over signatures, not argument values, because some
  # arguments may not have signatures.
  for arg_name in arg_signature_by_name:
    arg_signature = arg_signature_by_name[arg_name]
    arg_value = arg_value_by_name[arg_name]

    arg_type = arg_signature.annotation
    if not _is_tensor_type(arg_type):
      continue

    if not hasattr(arg_value, 'shape'):
      message = textwrap.dedent(f"""\
      Function '{func.__name__}': argument '{arg_name}' has type
      annotation '{arg_type}', but actual type is
      '{type(arg_value).__name__}'.
      """)
      message_one_line = message.replace('\n', ' ')
      raise TypeError(message_one_line)

    # If arg_type is Tensor2[Height, Width],
    # then arg_type.__args__ == (Height, Width).
    annotated_arg_rank = len(arg_type.__args__)
    actual_arg_rank = len(arg_value.shape)

    if annotated_arg_rank != actual_arg_rank:
      arg_name = arg_signature.name
      message = textwrap.dedent(f"""\
      Function '{func.__name__}': argument '{arg_name}' has type
      annotation '{arg_type}' with rank {annotated_arg_rank},
      but actual shape is '{arg_value.shape}' with rank {actual_arg_rank}
      """)
      message_one_line = message.replace('\n', ' ')
      raise TypeError(message_one_line)

  # ===== Call function. =====

  func_return_value = func(*args, **kwargs)

  # ===== Verify return. =====

  return_type = sig.return_annotation
  if not _is_tensor_type(return_type):
    return func_return_value

  if not hasattr(func_return_value, 'shape'):
    message = textwrap.dedent(f"""\
    Function '{func.__name__}': return has type annotation '{return_type}'
    but actual return type is '{type(func_return_value).__name__}'
    """)
    message_one_line = message.replace('\n', ' ')
    raise TypeError(message_one_line)

  annotated_rank = len(return_type.__args__)
  actual_rank = len(func_return_value.shape)
  if annotated_rank != actual_rank:
    message = textwrap.dedent(f"""\
    Function '{func.__name__}': return has type annotation '{return_type}'
    with rank {annotated_rank}, but actual shape is
    '{func_return_value.shape}' with rank {actual_rank}
    """)
    message_one_line = message.replace('\n', ' ')
    raise TypeError(message_one_line)

  return func_return_value


def _is_typed_tuple(x):
  return (
      hasattr(x, '__origin__')
      # `Tuple` for e.g. Python 3.6, `tuple` for e.g. Python 3.9.
      and x.__origin__ in (Tuple, tuple)
  )


def _is_typed_dict(x):
  return (
      hasattr(x, '__origin__')
      # `Dict` for e.g. Python 3.6, `dict` for e.g. Python 3.9.
      and x.__origin__ in (Dict, dict)
  )


def _is_typed_namedtuple(x):
  return hasattr(x, '_fields') and (
      hasattr(x, '_field_types')  # Python 3.6
      or hasattr(x, '__annotations__')  # Python 3.9
  )


def _tree_type_to_type_tree(tree_type: Any) -> Any:
  """Converts a tree-like type to a tree of the component types.

  Examples:

    T = Tuple[int, str]
    tree_type_to_type_tree(T) == (int, str)

    T2 = Dict[str, Tuple[float]]
    tree_type_to_type_tree(T2) == {str: (float,)}

    T3 = List[bool]
    tree_type_to_type_tree(T3) == [bool]

    class T3(NamedTuple):
      obses: Tuple[np.ndarray]
      actions: Tuple[np.ndarray]
    tree_type_to_type_tree(T3) == T3(obses=(np.ndarray,), actions=(np.ndarray,))

  If any of the items in the tree is unparameterised, it is not converted:

    T = Tuple[List, str]
    tree_type_to_type_tree(T) == (List, str)

  Args:
    tree_type: The tree-like type to convert.
  Returns:
    A tree of the component types.
  """
  def convert_tuple(x):
    if not _is_typed_tuple(x):
      return x
    # Check for unparameterised Tuple.
    if (
        not hasattr(x, '__args__') or  # Python 3.9
        x.__args__ is None or  # Python 3.6
        not x.__args__  # Python 3.7
    ):
      return x
    # If x is Tuple[int, str, float], x.__args__ will be (int, str, float).
    args = x.__args__
    # Check for Tuple[()].
    if args == ((),):
      return ()
    return args

  def convert_dict(x):
    if not _is_typed_dict(x):
      return x
    # Check for unparameterised Dict.
    if (
        not hasattr(x, '__args__') or  # Python 3.9
        x.__args__ is None or  # Python 3.6
        # Python 3.7
        x.__args__ == (typing.KT, typing.VT)  # pytype: disable=module-attr
    ):
      return x
    # If x is Dict[str, int], then x.__args__ should be (str, int).
    key_type, value_type = x.__args__
    return {key_type: value_type}

  def convert_named_tuple(x):
    if not _is_typed_namedtuple(x):
      return x
    try:
      # Python 3.6/3.7
      args = dict(x._field_types)  # pylint: disable=protected-access
    except AttributeError:
      # Python 3.9
      args = x.__annotations__
    return x(**args)

  type_tree = tree_type
  # Right now, `type_tree` doesn't even look like a tree.
  # So first, we have to try and convert the top-level type to a tree,
  # e.g. Tuple[Tuple[int]] -> (Tuple[int],)
  for f in (convert_tuple, convert_dict, convert_named_tuple):
    type_tree = f(type_tree)

  # Now we just have to keep converting elements of the tree until all
  # elements have been converted.
  prev_type_tree = copy.deepcopy(type_tree)
  while True:
    for f in (convert_tuple, convert_dict, convert_named_tuple):
      type_tree = tree.map_structure(f, type_tree)
    if type_tree == prev_type_tree:
      break
    prev_type_tree = type_tree

  return type_tree



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tools/templates.py."""

import unittest

from tensor_annotations.tools import templates


class TemplateTests(unittest.TestCase):

  def test_axis_list(self):
    self.assertEqual(
        templates.axis_list(n_axes=1),
        'A1'
    )
    self.assertEqual(
        templates.axis_list(n_axes=2),
        'A1, A2'
    )
    self.assertEqual(
        templates.axis_list(n_axes=1, reverse=True),
        'A1'
    )
    self.assertEqual(
        templates.axis_list(n_axes=2, reverse=True),
        'A2, A1'
    )

  def test_jax_array_type(self):
    self.assertEqual(
        'Array1[A1]',
        templates.jax_array_type(n_axes=1)
    )
    self.assertEqual(
        'Array2[A1, A2]',
        templates.jax_array_type(n_axes=2)
    )

  def test_transpose_axes(self):
    self.assertEqual(
        list(templates.transpose_axes(1)),
        [
            templates.TransposeAxes(
                n_axes=1,
                all_axes='[A1]',
                transpose_axes='Tuple[L0]',
                result_axes='[A1]'
            )
        ]
    )

    self.assertEqual(
        list(templates.transpose_axes(2)),
        [
            templates.TransposeAxes(
                n_axes=2,
                all_axes='[A1, A2]',
                transpose_axes='Tuple[L0, L1]',
                result_axes='[A1, A2]'
            ),
            templates.TransposeAxes(
                n_axes=2,
                all_axes='[A1, A2]',
                transpose_axes='Tuple[L1, L0]',
                result_axes='[A2, A1]'
            ),
        ]
    )

  def test_reduction_axes(self):
    self.assertEqual(
        list(templates.reduction_axes(1)),
        [
            templates.ReductionAxes(
                n_axes=1,
                all_axes='[A1]',
                reduction_axes='L0',
                remaining_n_axes=0,
                remaining_axes=''
            )
        ]
    )

    self.assertEqual(
        list(templates.reduction_axes(2)),
        [
            templates.ReductionAxes(
                n_axes=2,
                all_axes='[A1, A2]',
                reduction_axes='L0',
                remaining_n_axes=1,
                remaining_axes='[A2]'
            ),
            templates.ReductionAxes(
                n_axes=2,
                all_axes='[A1, A2]',
                reduction_axes='L1',
                remaining_n_axes=1,
                remaining_axes='[A1]'
            ),
            templates.ReductionAxes(
                n_axes=2,
                all_axes='[A1, A2]',
                reduction_axes='Tuple[L0, L1]',
                remaining_n_axes=0,
                remaining_axes=''
            ),
        ]
    )


if __name__ == '__main__':
  unittest.main()



"""Type stubs for custom TensorFlow tensor classes.

NOTE: This file is generated from templates/tensorflow_tensors.pyi.

To regenerate, run the following from the tensor_annotations directory:
   tools/render_tensor_template.py \
     --template templates/tensorflow_tensors.pyi \
     --out tensorflow.pyi
"""

from typing import Any, TypeVar, Tuple, Sequence, Generic, overload, Union

import tensorflow as tf
from tensor_annotations.axes import Axis


A1 = TypeVar('A1', bound=Axis)
A2 = TypeVar('A2', bound=Axis)
A3 = TypeVar('A3', bound=Axis)
A4 = TypeVar('A4', bound=Axis)
A5 = TypeVar('A5', bound=Axis)
A6 = TypeVar('A6', bound=Axis)
A7 = TypeVar('A7', bound=Axis)
A8 = TypeVar('A8', bound=Axis)

Number = Union[int, float]


# A quick refresher on broadcasting rules:
# 1. Tensor[A, B] + scalar = Tensor[A, B]
# 2. Otherwise, start with trailing dimension of each tensor and work
#    forwards. Broadcasting is possible if, for each axis, the dimensions
#    of that axis in each tensor are either a) equal or b) one of them is 1.
# We deliberately ignore case b) for the time being since we don't support
# literal shapes yet.

class Tensor0:
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor0: ...

  def __neg__(self) -> Tensor0: ...

  def __pos__(self) -> Tensor0: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __add__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __add__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __add__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __add__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __add__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __add__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __add__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __add__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __sub__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __sub__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __sub__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __sub__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __sub__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __sub__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __sub__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __sub__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __floordiv__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __floordiv__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __floordiv__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __floordiv__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __floordiv__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __floordiv__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __floordiv__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __floordiv__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __truediv__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __truediv__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __truediv__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __truediv__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __truediv__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __truediv__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __truediv__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __truediv__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __pow__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __pow__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __pow__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __pow__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __pow__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __pow__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __pow__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __pow__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __lt__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __lt__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __lt__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __lt__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __lt__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __lt__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __lt__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __lt__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __le__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __le__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __le__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __le__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __le__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __le__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __le__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __le__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __ge__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __ge__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __ge__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __ge__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __ge__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __ge__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __ge__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __ge__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __gt__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __gt__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __gt__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __gt__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __gt__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __gt__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __gt__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __gt__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __mul__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __mul__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __mul__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __mul__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __mul__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __mul__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __mul__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __mul__(self, other: Tensor8) -> Tensor8: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor0: ...

  @overload
  def __rmul__(self, other: Tensor1) -> Tensor1: ...

  @overload
  def __rmul__(self, other: Tensor2) -> Tensor2: ...

  @overload
  def __rmul__(self, other: Tensor3) -> Tensor3: ...

  @overload
  def __rmul__(self, other: Tensor4) -> Tensor4: ...

  @overload
  def __rmul__(self, other: Tensor5) -> Tensor5: ...

  @overload
  def __rmul__(self, other: Tensor6) -> Tensor6: ...

  @overload
  def __rmul__(self, other: Tensor7) -> Tensor7: ...

  @overload
  def __rmul__(self, other: Tensor8) -> Tensor8: ...

  # END: Binary element-wise operators


class Tensor1(Generic[A1]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor1[A1]: ...

  def __neg__(self) -> Tensor1[A1]: ...

  def __pos__(self) -> Tensor1[A1]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __add__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __sub__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __truediv__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __pow__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __lt__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __le__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __le__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __ge__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __gt__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __mul__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor1[A1]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor1[A1]: ...

  @overload
  def __rmul__(self, other: Tensor1[A1]) -> Tensor1[A1]: ...

  # END: Binary element-wise operators


class Tensor2(Generic[A1, A2]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor2[A1, A2]: ...

  def __neg__(self) -> Tensor2[A1, A2]: ...

  def __pos__(self) -> Tensor2[A1, A2]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __add__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __add__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __sub__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __sub__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __pow__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __pow__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __lt__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __lt__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __le__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __le__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __le__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __ge__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __ge__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __gt__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __gt__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __mul__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __mul__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Tensor1[A2]) -> Tensor2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Tensor2[A1, A2]) -> Tensor2[A1, A2]: ...

  # END: Binary element-wise operators


class Tensor3(Generic[A1, A2, A3]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor3[A1, A2, A3]: ...

  def __neg__(self) -> Tensor3[A1, A2, A3]: ...

  def __pos__(self) -> Tensor3[A1, A2, A3]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Tensor1[A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Tensor2[A2, A3]) -> Tensor3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Tensor3[A1, A2, A3]) -> Tensor3[A1, A2, A3]: ...

  # END: Binary element-wise operators


class Tensor4(Generic[A1, A2, A3, A4]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor4[A1, A2, A3, A4]: ...

  def __neg__(self) -> Tensor4[A1, A2, A3, A4]: ...

  def __pos__(self) -> Tensor4[A1, A2, A3, A4]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Tensor1[A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Tensor2[A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Tensor3[A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Tensor4[A1, A2, A3, A4]) -> Tensor4[A1, A2, A3, A4]: ...


class Tensor5(Generic[A1, A2, A3, A4, A5]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor5[A1, A2, A3, A4, A5]: ...

  def __neg__(self) -> Tensor5[A1, A2, A3, A4, A5]: ...

  def __pos__(self) -> Tensor5[A1, A2, A3, A4, A5]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Tensor1[A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Tensor2[A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Tensor3[A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Tensor4[A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Tensor5[A1, A2, A3, A4, A5]) -> Tensor5[A1, A2, A3, A4, A5]: ...

  # END: Binary element-wise operators


class Tensor6(Generic[A1, A2, A3, A4, A5, A6]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  def __neg__(self) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  def __pos__(self) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Tensor1[A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Tensor2[A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Tensor3[A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Tensor4[A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Tensor5[A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Tensor6[A1, A2, A3, A4, A5, A6]) -> Tensor6[A1, A2, A3, A4, A5, A6]: ...

  # END: Binary element-wise operators


class Tensor7(Generic[A1, A2, A3, A4, A5, A6, A7]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  def __neg__(self) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  def __pos__(self) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor1[A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor2[A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor3[A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor4[A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor5[A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor6[A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Tensor7[A1, A2, A3, A4, A5, A6, A7]) -> Tensor7[A1, A2, A3, A4, A5, A6, A7]: ...

  # END: Binary element-wise operators


class Tensor8(Generic[A1, A2, A3, A4, A5, A6, A7, A8]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: tf.TensorShape

  # BEGIN: Unary operators

  def __abs__(self) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  def __neg__(self) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  def __pos__(self) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Number) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor0) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor1[A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor2[A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor3[A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor4[A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor5[A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor6[A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor7[A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Tensor8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  # END: Binary element-wise operators
